package org.gemoc.execution.sequential.javaengine.ui.debug;

import com.google.common.base.Objects;
import fr.inria.diverse.trace.api.IStep;
import fr.inria.diverse.trace.api.IValueTrace;
import fr.inria.diverse.trace.gemoc.api.IGemocTraceManager;
import fr.inria.diverse.trace.gemoc.api.IMultiDimensionalTraceAddon;
import fr.obeo.dsl.debug.ide.event.IDSLDebugEventProcessor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.BiPredicate;
import java.util.function.Consumer;
import java.util.function.Supplier;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.xtext.naming.DefaultDeclarativeQualifiedNameProvider;
import org.eclipse.xtext.naming.QualifiedName;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.gemoc.execution.sequential.javaengine.ui.Activator;
import org.gemoc.execution.sequential.javaengine.ui.debug.GenericSequentialModelDebugger;
import org.gemoc.execution.sequential.javaengine.ui.debug.WrapperOmniscientDebugTimeLine;
import org.gemoc.executionframework.engine.core.AbstractSequentialExecutionEngine;
import org.gemoc.executionframework.engine.core.EngineStoppedException;
import org.gemoc.executionframework.engine.mse.MSE;
import org.gemoc.executionframework.engine.mse.MSEOccurrence;
import org.gemoc.xdsmlframework.api.core.IBasicExecutionEngine;
import org.gemoc.xdsmlframework.api.core.ISequentialExecutionEngine;

@SuppressWarnings("all")
public class OmniscientGenericSequentialModelDebugger extends GenericSequentialModelDebugger {
  private final IMultiDimensionalTraceAddon traceAddon;
  
  /**
   * -1 means we are on the last recorded state.
   * Otherwise value of the last jump index.
   */
  private int lastJumpIndex;
  
  private boolean inThePast = false;
  
  private int steppingOverStackFrameIndex = (-1);
  
  private int steppingReturnStackFrameIndex = (-1);
  
  public OmniscientGenericSequentialModelDebugger(final IDSLDebugEventProcessor target, final ISequentialExecutionEngine engine, final IMultiDimensionalTraceAddon addon) {
    super(target, engine);
    this.traceAddon = addon;
    WrapperOmniscientDebugTimeLine _wrapperOmniscientDebugTimeLine = new WrapperOmniscientDebugTimeLine(this);
    this.traceAddon.setTimeLineNotifier(_wrapperOmniscientDebugTimeLine);
    this.lastJumpIndex = (-1);
    this.callStack = Collections.EMPTY_LIST;
  }
  
  public boolean isInReplayMode() {
    return this.inThePast;
  }
  
  private MSE getMSEFromStep(final EObject caller, final IStep step) {
    MSE mse = null;
    if ((caller instanceof MSEOccurrence)) {
      MSE _mse = ((MSEOccurrence) caller).getMse();
      mse = _mse;
    } else {
      String _containingClassName = step.getContainingClassName();
      String _operationName = step.getOperationName();
      MSE _findOrCreateMSE = ((AbstractSequentialExecutionEngine) this.engine).findOrCreateMSE(caller, _containingClassName, _operationName);
      mse = _findOrCreateMSE;
    }
    return mse;
  }
  
  private void pushStackFrame(final String threadName, final IStep step) {
    EObject caller = null;
    MSE mse = null;
    String name = null;
    Map<String, Object> _parameters = step.getParameters();
    Set<Map.Entry<String, Object>> _entrySet = _parameters.entrySet();
    final Function1<Map.Entry<String, Object>, Boolean> _function = new Function1<Map.Entry<String, Object>, Boolean>() {
      @Override
      public Boolean apply(final Map.Entry<String, Object> es) {
        String _key = es.getKey();
        return Boolean.valueOf(_key.equals("caller"));
      }
    };
    final Map.Entry<String, Object> callerEntry = IterableExtensions.<Map.Entry<String, Object>>findFirst(_entrySet, _function);
    boolean _notEquals = (!Objects.equal(callerEntry, null));
    if (_notEquals) {
      Object _value = callerEntry.getValue();
      final EObject entryValue = ((EObject) _value);
      MSE _mSEFromStep = this.getMSEFromStep(entryValue, step);
      mse = _mSEFromStep;
      String _name = mse.getName();
      name = _name;
    } else {
      IStep _parentStep = step.getParentStep();
      Map<String, Object> _parameters_1 = _parentStep.getParameters();
      Object _get = _parameters_1.get("this");
      final EObject parentStep = ((EObject) _get);
      MSE _mSEFromStep_1 = this.getMSEFromStep(parentStep, step);
      mse = _mSEFromStep_1;
      String _name_1 = mse.getName();
      String _plus = (_name_1 + "_implicitStep");
      name = _plus;
    }
    EObject _caller = mse.getCaller();
    caller = _caller;
    EClass _eClass = caller.eClass();
    String _name_2 = _eClass.getName();
    String _plus_1 = (_name_2 + " (");
    String _plus_2 = (_plus_1 + name);
    String _plus_3 = (_plus_2 + ") [");
    String _string = caller.toString();
    String _plus_4 = (_plus_3 + _string);
    String _plus_5 = (_plus_4 + "]");
    name = _plus_5;
    final DefaultDeclarativeQualifiedNameProvider nameprovider = new DefaultDeclarativeQualifiedNameProvider();
    final QualifiedName qname = nameprovider.getFullyQualifiedName(caller);
    String _xifexpression = null;
    if ((qname != null)) {
      _xifexpression = qname.toString();
    } else {
      _xifexpression = caller.toString();
    }
    final String objectName = _xifexpression;
    EOperation _action = mse.getAction();
    String _name_3 = null;
    if (_action!=null) {
      _name_3=_action.getName();
    }
    final String opName = _name_3;
    EClass _eClass_1 = caller.eClass();
    final String callerType = _eClass_1.getName();
    final String prettyName = (((((("(" + callerType) + ") ") + objectName) + " -> ") + opName) + "()");
    this.pushStackFrame(threadName, prettyName, caller, caller);
  }
  
  private void loadLastState(final String threadName) {
    if ((this.lastJumpIndex != (-1))) {
      int _lastIndex = this.getLastIndex();
      this.jumpToState(_lastIndex);
    }
  }
  
  private List<Object> computeNewStack(final MSEOccurrence mseOccurrence) {
    final ArrayList<Object> newPath = new ArrayList<Object>();
    newPath.add(mseOccurrence);
    EObject container = mseOccurrence.eContainer();
    while (((!Objects.equal(container, null)) && (container instanceof MSEOccurrence))) {
      {
        newPath.add(0, container);
        EObject _eContainer = container.eContainer();
        container = _eContainer;
      }
    }
    return newPath;
  }
  
  @Override
  public void aboutToExecuteMSEOccurrence(final IBasicExecutionEngine executionEngine, final MSEOccurrence mseOccurrence) {
    List<Object> _computeNewStack = this.computeNewStack(mseOccurrence);
    this.newStack = _computeNewStack;
    boolean _control = this.control(this.threadName, mseOccurrence);
    boolean _not = (!_control);
    if (_not) {
      throw new EngineStoppedException("Debug thread has stopped.");
    }
  }
  
  @Override
  public void mseOccurrenceExecuted(final IBasicExecutionEngine engine, final MSEOccurrence mseOccurrence) {
    List<Object> _computeNewStack = this.computeNewStack(mseOccurrence);
    this.newStack = _computeNewStack;
  }
  
  @Override
  protected void updateStack(final String threadName, final EObject instruction) {
    if ((!this.inThePast)) {
      this.doStuff(this.newStack, true);
    }
  }
  
  @Override
  public void resume() {
    if ((!this.executionTerminated)) {
      if (this.inThePast) {
        this.loadLastState(this.threadName);
      }
      super.resume();
    }
  }
  
  @Override
  public void resume(final String threadName) {
    if ((!this.executionTerminated)) {
      if (this.inThePast) {
        this.loadLastState(threadName);
      }
      super.resume(threadName);
    }
  }
  
  @Override
  public void terminate() {
    super.terminate();
    Activator _default = Activator.getDefault();
    final Supplier<OmniscientGenericSequentialModelDebugger> _function = new Supplier<OmniscientGenericSequentialModelDebugger>() {
      @Override
      public OmniscientGenericSequentialModelDebugger get() {
        return null;
      }
    };
    _default.setDebuggerSupplier(_function);
  }
  
  @Override
  protected void setupStepOverPredicateBreak() {
    abstract class __OmniscientGenericSequentialModelDebugger_1 implements BiPredicate<IBasicExecutionEngine, MSEOccurrence> {
      MSEOccurrence steppedOver;
    }
    
    if ((this.steppingOverStackFrameIndex != (-1))) {
      final ISequentialExecutionEngine seqEngine = ((ISequentialExecutionEngine) this.engine);
      final Deque<MSEOccurrence> stack = seqEngine.getCurrentStack();
      int _size = stack.size();
      final int idx = (_size - this.steppingOverStackFrameIndex);
      __OmniscientGenericSequentialModelDebugger_1 ___OmniscientGenericSequentialModelDebugger_1 = new __OmniscientGenericSequentialModelDebugger_1() {
        {
          steppedOver = ((MSEOccurrence[])Conversions.unwrapArray(stack, MSEOccurrence.class))[idx];
        }
        @Override
        public boolean test(final IBasicExecutionEngine t, final MSEOccurrence u) {
          Deque<MSEOccurrence> _currentStack = seqEngine.getCurrentStack();
          boolean _contains = _currentStack.contains(this.steppedOver);
          return (!_contains);
        }
      };
      this.addPredicateBreak(___OmniscientGenericSequentialModelDebugger_1);
      this.steppingOverStackFrameIndex = (-1);
    } else {
      super.setupStepOverPredicateBreak();
    }
  }
  
  private IStep stepIntoResult;
  
  private IStep stepOverResult;
  
  private IStep stepReturnResult;
  
  private IStep backIntoResult;
  
  private IStep backOverResult;
  
  private IStep backOutResult;
  
  private List<IStep> callStack;
  
  private List<Object> newStack;
  
  private void doStuff(final List<Object> path, final boolean updateStack) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    IGemocTraceManager _traceManager_1 = this.traceAddon.getTraceManager();
    int _traceSize = _traceManager_1.getTraceSize();
    final List<IStep> rootSteps = _traceManager.getStepsForStates(0, _traceSize);
    final ArrayList<IStep> stepPath = new ArrayList<IStep>();
    final ArrayList<IStep> currentSteps = new ArrayList<IStep>();
    currentSteps.addAll(rootSteps);
    final Consumer<Object> _function = new Consumer<Object>() {
      @Override
      public void accept(final Object o) {
        final Function1<IStep, Boolean> _function = new Function1<IStep, Boolean>() {
          @Override
          public Boolean apply(final IStep s) {
            Map<String, Object> _parameters = s.getParameters();
            Object _get = _parameters.get("this");
            return Boolean.valueOf(Objects.equal(_get, o));
          }
        };
        final IStep step = IterableExtensions.<IStep>findFirst(currentSteps, _function);
        currentSteps.clear();
        boolean _notEquals = (!Objects.equal(step, null));
        if (_notEquals) {
          stepPath.add(step);
          List<IStep> _subSteps = step.getSubSteps();
          currentSteps.addAll(_subSteps);
        }
      }
    };
    path.forEach(_function);
    final List<IStep> stepPathUnmodifiable = Collections.<IStep>unmodifiableList(stepPath);
    IStep _computeStepInto = this.computeStepInto(stepPathUnmodifiable, rootSteps);
    this.stepIntoResult = _computeStepInto;
    IStep _computeStepOver = this.computeStepOver(stepPathUnmodifiable, rootSteps);
    this.stepOverResult = _computeStepOver;
    IStep _computeStepReturn = this.computeStepReturn(stepPathUnmodifiable, rootSteps);
    this.stepReturnResult = _computeStepReturn;
    IStep _computeBackInto = this.computeBackInto(stepPathUnmodifiable, rootSteps);
    this.backIntoResult = _computeBackInto;
    IStep _computeBackOver = this.computeBackOver(stepPathUnmodifiable, rootSteps);
    this.backOverResult = _computeBackOver;
    IStep _computeBackOut = this.computeBackOut(stepPathUnmodifiable, rootSteps);
    this.backOutResult = _computeBackOut;
    int i = 0;
    while ((((i < path.size()) && (i < this.callStack.size())) && Objects.equal(path.get(i), this.callStack.get(i)))) {
      i++;
    }
    if (updateStack) {
      for (int j = i; (j < this.callStack.size()); j++) {
        this.popStackFrame(this.threadName);
      }
      for (int j = i; (j < path.size()); j++) {
        IStep _get = stepPath.get(j);
        this.pushStackFrame(this.threadName, _get);
      }
    }
    this.callStack = stepPathUnmodifiable;
  }
  
  public IStep computeBackInto(final List<IStep> stepPath, final List<IStep> rootSteps) {
    IStep result = null;
    int _size = stepPath.size();
    boolean _greaterThan = (_size > 1);
    if (_greaterThan) {
      final List<IStep> reversedPath = ListExtensions.<IStep>reverseView(stepPath);
      final IStep currentStep = reversedPath.get(0);
      final IStep parentStep = reversedPath.get(1);
      final List<IStep> parentSubSteps = parentStep.getSubSteps();
      final int idx = parentSubSteps.indexOf(currentStep);
      if ((idx == 0)) {
        result = parentStep;
      } else {
        final IStep previousStep = parentSubSteps.get((idx - 1));
        IStep tmpStep = previousStep;
        List<IStep> tmpSubSteps = tmpStep.getSubSteps();
        while ((!tmpSubSteps.isEmpty())) {
          {
            IStep _last = IterableExtensions.<IStep>last(tmpSubSteps);
            tmpStep = _last;
            List<IStep> _subSteps = tmpStep.getSubSteps();
            tmpSubSteps = _subSteps;
          }
        }
        result = tmpStep;
      }
    } else {
      int _size_1 = stepPath.size();
      boolean _equals = (_size_1 == 1);
      if (_equals) {
        final IStep currentStep_1 = stepPath.get(0);
        final int idx_1 = rootSteps.indexOf(currentStep_1);
        if ((idx_1 > 0)) {
          final IStep previousStep_1 = rootSteps.get((idx_1 - 1));
          IStep tmpStep_1 = previousStep_1;
          List<IStep> tmpSubSteps_1 = tmpStep_1.getSubSteps();
          while ((!tmpSubSteps_1.isEmpty())) {
            {
              IStep _last = IterableExtensions.<IStep>last(tmpSubSteps_1);
              tmpStep_1 = _last;
              List<IStep> _subSteps = tmpStep_1.getSubSteps();
              tmpSubSteps_1 = _subSteps;
            }
          }
          result = tmpStep_1;
        }
      }
    }
    return result;
  }
  
  public IStep computeBackOver(final List<IStep> stepPath, final List<IStep> rootSteps) {
    boolean _isEmpty = stepPath.isEmpty();
    boolean _not = (!_isEmpty);
    if (_not) {
      final List<IStep> reversedPath = ListExtensions.<IStep>reverseView(stepPath);
      IStep _get = reversedPath.get(0);
      return this.findPreviousStep(reversedPath, rootSteps, _get, 1);
    }
    return null;
  }
  
  public IStep computeBackOut(final List<IStep> stepPath, final List<IStep> rootSteps) {
    int _size = stepPath.size();
    boolean _greaterThan = (_size > 1);
    if (_greaterThan) {
      final List<IStep> reversedPath = ListExtensions.<IStep>reverseView(stepPath);
      IStep _get = reversedPath.get(1);
      return this.findPreviousStep(reversedPath, rootSteps, _get, 2);
    }
    return null;
  }
  
  public IStep findPreviousStep(final List<IStep> stepPath, final List<IStep> rootSteps, final IStep previousStep, final int start) {
    IStep result = null;
    int i = start;
    IStep previous = previousStep;
    while ((Objects.equal(result, null) && (i < stepPath.size()))) {
      {
        final IStep currentStep = stepPath.get(i);
        final List<IStep> currentSubSteps = currentStep.getSubSteps();
        int _indexOf = currentSubSteps.indexOf(previous);
        int idx = (_indexOf - 1);
        if ((idx > 0)) {
          IStep _get = currentSubSteps.get(idx);
          result = _get;
        } else {
          previous = currentStep;
        }
        i++;
      }
    }
    boolean _equals = Objects.equal(result, null);
    if (_equals) {
      int _indexOf = rootSteps.indexOf(previous);
      final int idx = (_indexOf - 1);
      if ((idx > 0)) {
        IStep _get = rootSteps.get(idx);
        result = _get;
      }
    }
    return result;
  }
  
  public IStep findNextStep(final List<IStep> stepPath, final List<IStep> rootSteps, final IStep previousStep, final int start) {
    IStep result = null;
    int i = start;
    IStep previous = previousStep;
    while ((Objects.equal(result, null) && (i < stepPath.size()))) {
      {
        final IStep currentStep = stepPath.get(i);
        final List<IStep> currentSubSteps = currentStep.getSubSteps();
        boolean _isEmpty = currentSubSteps.isEmpty();
        if (_isEmpty) {
          previous = currentStep;
        } else {
          boolean _equals = Objects.equal(previous, null);
          if (_equals) {
            IStep _get = currentSubSteps.get(0);
            result = _get;
          } else {
            int _indexOf = currentSubSteps.indexOf(previous);
            final int idx = (_indexOf + 1);
            int _size = currentSubSteps.size();
            boolean _lessThan = (idx < _size);
            if (_lessThan) {
              IStep _get_1 = currentSubSteps.get(idx);
              result = _get_1;
            } else {
              previous = currentStep;
            }
          }
        }
        i++;
      }
    }
    boolean _equals = Objects.equal(result, null);
    if (_equals) {
      int _indexOf = rootSteps.indexOf(previous);
      final int idx = (_indexOf + 1);
      int _size = rootSteps.size();
      boolean _lessThan = (idx < _size);
      if (_lessThan) {
        IStep _get = rootSteps.get(idx);
        result = _get;
      }
    }
    return result;
  }
  
  public IStep computeStepInto(final List<IStep> stepPath, final List<IStep> rootSteps) {
    List<IStep> _reverseView = ListExtensions.<IStep>reverseView(stepPath);
    return this.findNextStep(_reverseView, rootSteps, null, 0);
  }
  
  public IStep computeStepOver(final List<IStep> stepPath, final List<IStep> rootSteps) {
    boolean _isEmpty = stepPath.isEmpty();
    boolean _not = (!_isEmpty);
    if (_not) {
      final List<IStep> reversedPath = ListExtensions.<IStep>reverseView(stepPath);
      IStep _get = reversedPath.get(0);
      return this.findNextStep(reversedPath, rootSteps, _get, 1);
    }
    return null;
  }
  
  public IStep computeStepReturn(final List<IStep> stepPath, final List<IStep> rootSteps) {
    int _size = stepPath.size();
    boolean _greaterThan = (_size > 1);
    if (_greaterThan) {
      final List<IStep> reversedPath = ListExtensions.<IStep>reverseView(stepPath);
      IStep _get = reversedPath.get(1);
      return this.findNextStep(reversedPath, rootSteps, _get, 2);
    }
    return null;
  }
  
  @Override
  public void stepInto(final String threadName) {
    if (this.inThePast) {
      boolean _equals = Objects.equal(this.stepIntoResult, null);
      if (_equals) {
        if ((this.lastJumpIndex == (-1))) {
          this.inThePast = false;
          super.stepInto(threadName);
        } else {
          throw new IllegalStateException("Next step was not found");
        }
      } else {
        this.jumpBeforeStep(this.stepIntoResult);
      }
    } else {
      super.stepInto(threadName);
    }
  }
  
  @Override
  public void stepOver(final String threadName) {
    if (this.inThePast) {
      boolean _equals = Objects.equal(this.stepOverResult, null);
      if (_equals) {
        IStep _last = IterableExtensions.<IStep>last(this.callStack);
        int _endingIndex = _last.getEndingIndex();
        boolean _equals_1 = (_endingIndex == (-1));
        if (_equals_1) {
          this.steppingOverStackFrameIndex = (this.nbStackFrames - 1);
          this.loadLastState(threadName);
          this.inThePast = false;
          super.stepOver(threadName);
        } else {
          throw new IllegalStateException("Next step was not found");
        }
      } else {
        this.jumpBeforeStep(this.stepOverResult);
      }
    } else {
      super.stepOver(threadName);
    }
  }
  
  @Override
  protected void setupStepReturnPredicateBreak() {
    abstract class __OmniscientGenericSequentialModelDebugger_2 implements BiPredicate<IBasicExecutionEngine, MSEOccurrence> {
      MSEOccurrence steppedReturn;
    }
    
    if ((this.steppingReturnStackFrameIndex != (-1))) {
      final ISequentialExecutionEngine seqEngine = ((ISequentialExecutionEngine) this.engine);
      final Deque<MSEOccurrence> stack = seqEngine.getCurrentStack();
      int _size = stack.size();
      final int idx = (_size - this.steppingReturnStackFrameIndex);
      __OmniscientGenericSequentialModelDebugger_2 ___OmniscientGenericSequentialModelDebugger_2 = new __OmniscientGenericSequentialModelDebugger_2() {
        {
          steppedReturn = ((MSEOccurrence[])Conversions.unwrapArray(stack, MSEOccurrence.class))[idx];
        }
        @Override
        public boolean test(final IBasicExecutionEngine t, final MSEOccurrence u) {
          Deque<MSEOccurrence> _currentStack = seqEngine.getCurrentStack();
          boolean _contains = _currentStack.contains(this.steppedReturn);
          return (!_contains);
        }
      };
      this.addPredicateBreak(___OmniscientGenericSequentialModelDebugger_2);
      this.steppingReturnStackFrameIndex = (-1);
    } else {
      super.setupStepReturnPredicateBreak();
    }
  }
  
  @Override
  public void stepReturn(final String threadName) {
    if (this.inThePast) {
      boolean _equals = Objects.equal(this.stepReturnResult, null);
      if (_equals) {
        boolean _and = false;
        int _size = this.callStack.size();
        boolean _greaterThan = (_size > 1);
        if (!_greaterThan) {
          _and = false;
        } else {
          int _size_1 = this.callStack.size();
          int _minus = (_size_1 - 2);
          IStep _get = this.callStack.get(_minus);
          int _endingIndex = _get.getEndingIndex();
          boolean _equals_1 = (_endingIndex == (-1));
          _and = _equals_1;
        }
        if (_and) {
          this.steppingReturnStackFrameIndex = (this.nbStackFrames - 2);
          this.loadLastState(threadName);
          this.inThePast = false;
          super.stepReturn(threadName);
        } else {
          throw new IllegalStateException("Next step was not found");
        }
      } else {
        this.jumpBeforeStep(this.stepReturnResult);
      }
    } else {
      super.stepReturn(threadName);
    }
  }
  
  public boolean canStepBackInto(final String threadName) {
    return (!Objects.equal(this.backIntoResult, null));
  }
  
  public boolean canStepBackOver(final String threadName) {
    return (!Objects.equal(this.backOverResult, null));
  }
  
  public boolean canStepBackOut(final String threadName) {
    return (!Objects.equal(this.backOutResult, null));
  }
  
  public void stepBackInto(final String threadName) {
    boolean _notEquals = (!Objects.equal(this.backIntoResult, null));
    if (_notEquals) {
      this.inThePast = true;
      this.jumpBeforeStep(this.backIntoResult);
    }
  }
  
  public void stepBackOver(final String threadName) {
    boolean _notEquals = (!Objects.equal(this.backOverResult, null));
    if (_notEquals) {
      this.inThePast = true;
      this.jumpBeforeStep(this.backOverResult);
    }
  }
  
  public void stepBackOut(final String threadName) {
    boolean _notEquals = (!Objects.equal(this.backOutResult, null));
    if (_notEquals) {
      this.inThePast = true;
      this.jumpBeforeStep(this.backOutResult);
    }
  }
  
  public boolean canStepValue(final int traceIndex) {
    return true;
  }
  
  private int getNextValueIndex(final IValueTrace valueTrace) {
    int stateIndex = this.getCurrentStateIndex();
    final int currentValueIndex = valueTrace.getActiveValueIndex(stateIndex);
    int valueIndex = currentValueIndex;
    while (((stateIndex < this.getLastIndex()) && ((valueIndex == currentValueIndex) || (valueIndex == (-1))))) {
      {
        stateIndex++;
        int _activeValueIndex = valueTrace.getActiveValueIndex(stateIndex);
        valueIndex = _activeValueIndex;
      }
    }
    if (((valueIndex == currentValueIndex) || (valueIndex == (-1)))) {
      return valueTrace.getSize();
    }
    return valueIndex;
  }
  
  public void stepValue(final int traceIndex) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    List<IValueTrace> _allValueTraces = _traceManager.getAllValueTraces();
    final IValueTrace valueTrace = _allValueTraces.get(traceIndex);
    final int i = this.getNextValueIndex(valueTrace);
    boolean _and = false;
    int _size = valueTrace.getSize();
    boolean _lessThan = (i < _size);
    if (!_lessThan) {
      _and = false;
    } else {
      _and = (i != (-1));
    }
    if (_and) {
      EObject _value = valueTrace.getValue(i);
      this.jump(_value);
    } else {
      this.setupStepValuePredicateBreak(valueTrace, i);
      this.resume();
    }
  }
  
  private void setupStepValuePredicateBreak(final IValueTrace valueTrace, final int valueIndex) {
    this.addPredicateBreak(new BiPredicate<IBasicExecutionEngine, MSEOccurrence>() {
      @Override
      public boolean test(final IBasicExecutionEngine t, final MSEOccurrence u) {
        int _currentStateIndex = OmniscientGenericSequentialModelDebugger.this.getCurrentStateIndex();
        final int i = valueTrace.getActiveValueIndex(_currentStateIndex);
        final int j = valueIndex;
        return (i == j);
      }
    });
  }
  
  public boolean canBackValue(final int traceIndex) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    final List<IValueTrace> allValueTraces = _traceManager.getAllValueTraces();
    boolean _and = false;
    int _size = allValueTraces.size();
    boolean _lessThan = (traceIndex < _size);
    if (!_lessThan) {
      _and = false;
    } else {
      _and = (traceIndex > (-1));
    }
    if (_and) {
      final IValueTrace valueTrace = allValueTraces.get(traceIndex);
      int _currentStateIndex = this.getCurrentStateIndex();
      final int currentValueIndex = valueTrace.getActiveValueIndex(_currentStateIndex);
      int stateIndex = this.getCurrentStateIndex();
      int valueIndex = valueTrace.getActiveValueIndex(stateIndex);
      while (((stateIndex > 0) && ((valueIndex == currentValueIndex) || (valueIndex == (-1))))) {
        {
          stateIndex--;
          int _activeValueIndex = valueTrace.getActiveValueIndex(stateIndex);
          valueIndex = _activeValueIndex;
        }
      }
      return ((valueIndex != currentValueIndex) && (valueIndex != (-1)));
    }
    return false;
  }
  
  private int getPreviousValueIndex(final IValueTrace valueTrace) {
    int _currentStateIndex = this.getCurrentStateIndex();
    final int currentValueIndex = valueTrace.getActiveValueIndex(_currentStateIndex);
    int _currentStateIndex_1 = this.getCurrentStateIndex();
    int stateIndex = (_currentStateIndex_1 - 1);
    int valueIndex = valueTrace.getActiveValueIndex(stateIndex);
    while (((stateIndex > 0) && ((valueIndex == currentValueIndex) || (valueIndex == (-1))))) {
      {
        stateIndex--;
        int _activeValueIndex = valueTrace.getActiveValueIndex(stateIndex);
        valueIndex = _activeValueIndex;
      }
    }
    return valueIndex;
  }
  
  public void backValue(final int traceIndex) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    List<IValueTrace> _allValueTraces = _traceManager.getAllValueTraces();
    final IValueTrace valueTrace = _allValueTraces.get(traceIndex);
    int _previousValueIndex = this.getPreviousValueIndex(valueTrace);
    EObject _value = valueTrace.getValue(_previousValueIndex);
    this.jump(_value);
  }
  
  /**
   * Returns the index of the last state in the trace.
   */
  private int getLastIndex() {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _traceSize = _traceManager.getTraceSize();
    return (_traceSize - 1);
  }
  
  /**
   * If we are in the past, we read the last jump index.
   * Otherwise, we read the trace size.
   */
  public int getCurrentStateIndex() {
    if ((this.lastJumpIndex != (-1))) {
      return this.lastJumpIndex;
    } else {
      return this.getLastIndex();
    }
  }
  
  private void jumpBeforeStep(final IStep step) {
    int _startingIndex = step.getStartingIndex();
    this.jumpToState(_startingIndex);
    final ArrayList<Object> newPath = new ArrayList<Object>();
    Map<String, Object> _parameters = step.getParameters();
    Object _get = _parameters.get("this");
    newPath.add(_get);
    IStep parent = step.getParentStep();
    while ((!Objects.equal(parent, null))) {
      {
        Map<String, Object> _parameters_1 = parent.getParameters();
        Object _get_1 = _parameters_1.get("this");
        newPath.add(0, _get_1);
        IStep _parentStep = parent.getParentStep();
        parent = _parentStep;
      }
    }
    this.doStuff(newPath, true);
  }
  
  private void jumpToState(final int i) {
    this.inThePast = true;
    int _lastIndex = this.getLastIndex();
    boolean _equals = (i == _lastIndex);
    if (_equals) {
      this.lastJumpIndex = (-1);
    } else {
      this.lastJumpIndex = i;
    }
    this.traceAddon.goTo(i);
    this.updateData(this.threadName, null);
    this.scheduleSelectLastStackframe(500);
  }
  
  /**
   * To be used by the timeline
   */
  public void jump(final int i) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _traceSize = _traceManager.getTraceSize();
    boolean _lessThan = (i < _traceSize);
    if (_lessThan) {
      IGemocTraceManager _traceManager_1 = this.traceAddon.getTraceManager();
      final List<IStep> rootSteps = _traceManager_1.getStepsForStates(i, i);
      final List<IStep> searchPath = new ArrayList<IStep>();
      IStep firstStateStep = null;
      boolean _isEmpty = rootSteps.isEmpty();
      boolean _not = (!_isEmpty);
      if (_not) {
        IStep currentStep = rootSteps.get(0);
        List<IStep> siblingSteps = rootSteps;
        while (Objects.equal(firstStateStep, null)) {
          boolean _and = false;
          int _startingIndex = currentStep.getStartingIndex();
          boolean _lessThan_1 = (_startingIndex < i);
          if (!_lessThan_1) {
            _and = false;
          } else {
            boolean _or = false;
            int _endingIndex = currentStep.getEndingIndex();
            boolean _greaterThan = (_endingIndex > i);
            if (_greaterThan) {
              _or = true;
            } else {
              int _endingIndex_1 = currentStep.getEndingIndex();
              boolean _equals = (_endingIndex_1 == (-1));
              _or = _equals;
            }
            _and = _or;
          }
          if (_and) {
            List<IStep> _subSteps = currentStep.getSubSteps();
            boolean _isEmpty_1 = _subSteps.isEmpty();
            if (_isEmpty_1) {
              throw new IllegalStateException("Unreachable state");
            } else {
              searchPath.add(0, currentStep);
              List<IStep> _subSteps_1 = currentStep.getSubSteps();
              siblingSteps = _subSteps_1;
              IStep _get = siblingSteps.get(0);
              currentStep = _get;
            }
          } else {
            boolean _and_1 = false;
            int _endingIndex_2 = currentStep.getEndingIndex();
            boolean _equals_1 = (_endingIndex_2 == i);
            if (!_equals_1) {
              _and_1 = false;
            } else {
              int _startingIndex_1 = currentStep.getStartingIndex();
              boolean _notEquals = (_startingIndex_1 != i);
              _and_1 = _notEquals;
            }
            if (_and_1) {
              List<IStep> _subSteps_2 = currentStep.getSubSteps();
              boolean _isEmpty_2 = _subSteps_2.isEmpty();
              if (_isEmpty_2) {
                IStep tmp = currentStep;
                currentStep = null;
                while (Objects.equal(currentStep, null)) {
                  {
                    int _indexOf = siblingSteps.indexOf(tmp);
                    final int idx = (_indexOf + 1);
                    int _size = siblingSteps.size();
                    boolean _lessThan_2 = (idx < _size);
                    if (_lessThan_2) {
                      IStep _get_1 = siblingSteps.get(idx);
                      currentStep = _get_1;
                    } else {
                      boolean _isEmpty_3 = searchPath.isEmpty();
                      if (_isEmpty_3) {
                        throw new IllegalStateException("Unreachable state");
                      } else {
                        IStep _remove = searchPath.remove(0);
                        tmp = _remove;
                        boolean _isEmpty_4 = searchPath.isEmpty();
                        if (_isEmpty_4) {
                          siblingSteps = rootSteps;
                        } else {
                          IStep _get_2 = searchPath.get(0);
                          List<IStep> _subSteps_3 = _get_2.getSubSteps();
                          siblingSteps = _subSteps_3;
                        }
                      }
                    }
                  }
                }
              } else {
                searchPath.add(0, currentStep);
                List<IStep> _subSteps_3 = currentStep.getSubSteps();
                siblingSteps = _subSteps_3;
                IStep _get_1 = siblingSteps.get(0);
                currentStep = _get_1;
              }
            } else {
              int _startingIndex_2 = currentStep.getStartingIndex();
              boolean _equals_2 = (_startingIndex_2 == i);
              if (_equals_2) {
                firstStateStep = currentStep;
              }
            }
          }
        }
      }
      this.jumpBeforeStep(firstStateStep);
    }
    this.scheduleSelectLastStackframe(500);
  }
  
  /**
   * To be used by the timeline
   */
  public void jump(final EObject o) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _stateOrValueIndex = _traceManager.getStateOrValueIndex(o);
    this.jump(_stateOrValueIndex);
  }
  
  @Override
  public boolean validateVariableValue(final String threadName, final String variableName, final String value) {
    if (this.inThePast) {
      Status _status = new Status(IStatus.ERROR, Activator.PLUGIN_ID, "Illegal variable value set");
      ErrorDialog.openError(null, "Illegal variable value set", "Cannot set the value of a variable when in replay mode", _status);
      return false;
    }
    return super.validateVariableValue(threadName, variableName, value);
  }
  
  @Override
  public void engineStarted(final IBasicExecutionEngine executionEngine) {
    final Activator activator = Activator.getDefault();
    final Supplier<OmniscientGenericSequentialModelDebugger> _function = new Supplier<OmniscientGenericSequentialModelDebugger>() {
      @Override
      public OmniscientGenericSequentialModelDebugger get() {
        return OmniscientGenericSequentialModelDebugger.this;
      }
    };
    activator.setDebuggerSupplier(_function);
    super.engineStarted(executionEngine);
  }
  
  @Override
  public void engineAboutToStop(final IBasicExecutionEngine engine) {
    super.engineAboutToStop(engine);
    final IStep step = null;
    this.jumpBeforeStep(step);
  }
  
  @Override
  public void engineStopped(final IBasicExecutionEngine executionEngine) {
    final Activator activator = Activator.getDefault();
    activator.setDebuggerSupplier(null);
    super.engineStopped(executionEngine);
  }
}
